/*****************
* LIST FUNCTIONS *
*****************/

//
// Initialize a list.
//
void ListInitialize(ListHeader *List, int type)
{
  List->first = List->last = NULL;
  InitializeLock(&List->lock);
  List->type = type;
}

//
// Find the list element that preceeds the given one.
//
ListElementHeader *ListPrevious(ListElementHeader *elementheader)
{
  ListElementHeader *Current;

  Current = elementheader;
  while(Current->next != elementheader)
    Current = Current->next;

  return(Current);
}

//
// Insert a new element into a list.
//
void ListInsert(ListHeader *header, ListElementHeader *elementheader)
{
  // Enter critical region for list.
  Lock(&header->lock);

  if(header->first == NULL) {
    header->first = header->last = elementheader->next = elementheader;
  }
  else {
    ListElementHeader *Current;
    switch(header->type) {
    case PriorityListType:
      Current=header->first;
      if(elementheader->priority > Current->priority)
        header->first = elementheader;
      else
        while((elementheader->priority < Current->priority) &&
              ((Current = Current->next) != header->first));

      // Alter list header if new element ended up on top.
      if(Current == header->first)
        header->first = elementheader;

      // Point previous element in list to the new one.
      ListPrevious(Current)->next = elementheader;

      // Point new element to the one it's preceeding.
      elementheader->next = Current;
      break;
    case FIFOListType:
      elementheader->next = header->first;
      header->last->next = elementheader;
      header->last = elementheader;
      break;
    case LIFOListType:
      elementheader->next = header->first;
      header->last->next = elementheader;
      header->first = elementheader;
      break;
    }
  }

  // Exit critical region for list.
  Unlock(&header->lock);
}

//
// Remove an element from a list.
//
void ListRemove(ListHeader *header, ListElementHeader *elementheader)
{
  // Enter critical region for list.
  Lock(&header->lock);

  if(elementheader->next == elementheader)
    // Element is only one in list, clear list pointers.
    header->first = header->last = NULL;
  else {
    ListPrevious(elementheader)->next = elementheader->next;
    // If element is first in list change list->first to the next element in the list.
    if(header->first == elementheader)
      header->first = elementheader->next;
    // If element is last in list change list->last to the next element in the list.
    if(header->last == elementheader)
      header->last = elementheader->next;
  }

  // Exit critical region for list.
  Unlock(&header->lock);
}

//
// Find an element in a list by its ID.
//
ListElementHeader *ListFind(ListHeader *header, WORD idstring)
{
  ListElementHeader *Current;

  // Enter critical region for list.
  Lock(&header->lock);

  Current = header->first;
  if(Current->id != idstring)
    while(((Current = Current->next) != header->first) && (Current->id != idstring));
  if(Current->id != idstring)
    Current = NULL;

  // Exit critical region for list.
  Unlock(&header->lock);
  return(Current);
}

//
// Get the size of a list.
//
int ListSize(ListHeader *header)
{
  int size = 0;
  ListElementHeader *Current;

  // Enter critical region for list.
  Lock(&header->lock);

  Current = header->first;
  if(Current != NULL)
    do {size++;} while((Current = Current->next) != header->first);

  // Exit critical region for list.
  Unlock(&header->lock);
  return(size);
}
